import torch
import torch.nn as nn
import torch.nn.functional as F
import statistics
import torchvision.datasets as dsets
import general_energy_nets.predictive_coding as pc
import torchvision
import torch.optim as optim
import argparse
import matplotlib
import matplotlib.pyplot as plt
import torchvision.transforms as transforms
from tqdm import tqdm
from torchvision import datasets
from torch.utils.data import Dataset, DataLoader, TensorDataset
from torchvision.utils import save_image
from torch.autograd import Variable
#import Utils as u
import numpy as np
from tabulate import tabulate

# Parameters of the model
input_dim = 28*28
hidden_dim = 4096
output_dim = 10

# Number of iterations performed by the train and test inference, respectively.
epochs = 10
num_iterations = 10#5000
iterations_test = 250#2000

# Parameters of the algorithm. 
p_lr = 1e-5
x_lr = 0.5
xlr_test = 0.1
dataset_name = 'MNIST'
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
partial = 0 #500
batch_size = 500
test_batch_size = 2000

kt = 0
def generation(prediction_trainer):

    def callback_generation(t,pc_trainer):
        classes = torch.eye(output_dim).cuda()
        model.PClabel.get_x().data.copy_(classes)
        inputs = model.PCimg.get_x().data

    callback_generation_kwargs={'pc_trainer':prediction_trainer}
    fake_inputs = torch.rand(output_dim,784).cuda()
    prediction_trainer.train_on_batch(inputs=fake_inputs,is_log_progress=True,callback_after_t=callback_generation,callback_after_t_kwargs=callback_generation_kwargs)
    generated_images = model.PCimg.get_x().data.view(output_dim,1,28,28)[:10]
    save_image(generated_images.cpu(), f"epoch={epoch} - bi_cpc_MNIST_generated_images.png", nrow=10)

def forward_acc(model, test_loader):
    correct = 0.0
    total = 0.0
    pred_matrix_per_digit = np.zeros((10,10))

    for images, labels in test_loader:
        images = Variable(images.view(-1,28*28).cuda())
        outputs = model(images)
        _, predicted = torch.max(outputs.data, 1)
        total += labels.size(0)
        correct += (predicted.cpu() == labels.cpu()).sum()
        pred_matrix_per_digit +=accuracy_for_each_digit(predicted,labels)

    accuracy = int(10000 * correct / total)/100
    return accuracy, pred_matrix_per_digit

def accuracy_for_each_digit(predicted, labels):
    times = np.zeros((10,10))
    for i in range(0, 10):
        mask = (labels.cpu()==i)
        data_labeled_i = torch.masked_select(predicted,mask)
        for j in range(0,10):
            mask2 = data_labeled_i.cpu()==j
            times[i][j] = int(mask2.sum())
    return times.astype(int)
###########################################################################################

class Bidirectional_Model(nn.Module):
    def __init__(self, input_dim: int, hidden_dim: int, output_dim:int) -> None:
        super(Bidirectional_Model, self).__init__()

        self.PCimg = pc.PCLayer()
        self.PClabel = pc.PCLayer()

        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, hidden_dim)
        self.fc4 = nn.Linear(hidden_dim, output_dim)
        self.fc4inv = nn.Linear(output_dim, hidden_dim)
        self.fc3inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc2inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc1inv = nn.Linear(hidden_dim, input_dim)

        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.relu3 = nn.ReLU()
        self.relu4 = nn.ReLU()

        self.PClayer1 = pc.PCLayer() 
        self.PClayer2 = pc.PCLayer()
        self.PClayer3 = pc.PCLayer()

    def forward(self, x):
        out = self.PCimg(x)

        out = self.fc1(out)
        out = self.relu1(out)
        out = self.PClayer1(out)

        out = self.fc2(out)
        out = self.relu2(out)
        out = self.PClayer2(out)
        
        out = self.fc3(out)
        out = self.relu3(out)
        out = self.PClayer3(out)

        label = self.fc4(out)
        out = self.PClabel(label)

        out = self.fc4inv(out)
        out = self.PClayer3(out)
        out = self.relu3(out)
        out = self.fc3inv(out)

        out = self.PClayer2(out)
        out = self.relu2(out)
        out = self.fc2inv(out)

        out = self.PClayer1(out)
        out = self.relu1(out)      
        out = self.fc1inv(out)
        out = self.PCimg(out)

        return label

########################################################################
class Bidirectional_Model_L_1(nn.Module):
    def __init__(self, input_dim: int, hidden_dim: int, output_dim:int) -> None:
        super(Bidirectional_Model_L_1, self).__init__()

        self.PCimg = pc.PCLayer()
        self.PClabel = pc.PCLayer()

        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc4 = nn.Linear(hidden_dim, output_dim)
        self.fc4inv = nn.Linear(output_dim, hidden_dim)
        self.fc2inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc1inv = nn.Linear(hidden_dim, input_dim)

        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.relu4 = nn.ReLU()

        self.PClayer1 = pc.PCLayer() #PC Layer is what allows us to run inference
        self.PClayer2 = pc.PCLayer()

    def forward(self, x):
        out = self.PCimg(x)

        out = self.fc1(out)
        out = self.relu1(out)
        out = self.PClayer1(out)

        out = self.fc2(out)
        out = self.relu2(out)
        out = self.PClayer2(out)
        label = self.fc4(out)
        out = self.PClabel(label)

        out = self.fc4inv(out)

        out = self.PClayer2(out)
        out = self.relu2(out)
        out = self.fc2inv(out)

        out = self.PClayer1(out)
        out = self.relu1(out)      
        out = self.fc1inv(out)
        out = self.PCimg(out)

        return label

########################################################################
class Bidirectional_Model_L_3(nn.Module):
    def __init__(self, input_dim: int, hidden_dim: int, output_dim:int) -> None:
        super(Bidirectional_Model_L_3, self).__init__()

        self.PCimg = pc.PCLayer()
        self.PClabel = pc.PCLayer()

        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3_ = nn.Linear(hidden_dim, hidden_dim)
        self.fc4 = nn.Linear(hidden_dim, output_dim)
        self.fc4inv = nn.Linear(output_dim, hidden_dim)
        self.fc3inv_ = nn.Linear(hidden_dim, hidden_dim)
        self.fc3inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc2inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc1inv = nn.Linear(hidden_dim, input_dim)

        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.relu3 = nn.ReLU()
        self.relu3_ = nn.ReLU()
        self.relu4 = nn.ReLU()

        self.PClayer1 = pc.PCLayer() 
        self.PClayer2 = pc.PCLayer()
        self.PClayer3 = pc.PCLayer()
        self.PClayer3_ = pc.PCLayer()

    def forward(self, x):
        out = self.PCimg(x)

        out = self.fc1(out)
        out = self.relu1(out)
        out = self.PClayer1(out)

        out = self.fc2(out)
        out = self.relu2(out)
        out = self.PClayer2(out)
        
        out = self.fc3(out)
        out = self.relu3(out)
        out = self.PClayer3(out)

        out = self.fc3_(out)
        out = self.relu3_(out)
        out = self.PClayer3_(out)

        label = self.fc4(out)
        out = self.PClabel(label)

        out = self.fc4inv(out)
        out = self.PClayer3(out)
        out = self.relu3(out)
        out = self.fc3inv(out)

        out = self.PClayer3_(out)
        out = self.relu3_(out)
        out = self.fc3inv_(out)

        out = self.PClayer2(out)
        out = self.relu2(out)
        out = self.fc2inv(out)

        out = self.PClayer1(out)
        out = self.relu1(out)      
        out = self.fc1inv(out)
        out = self.PCimg(out)

        return label

########################################################################

class Bidirectional_Model_L_4(nn.Module):
    def __init__(self, input_dim: int, hidden_dim: int, output_dim:int) -> None:
        super(Bidirectional_Model_L_4, self).__init__()

        self.PCimg = pc.PCLayer()
        self.PClabel = pc.PCLayer()

        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.fc2_ = nn.Linear(hidden_dim, hidden_dim)
        self.fc3 = nn.Linear(hidden_dim, hidden_dim)
        self.fc3_ = nn.Linear(hidden_dim, hidden_dim)
        self.fc4 = nn.Linear(hidden_dim, output_dim)
        self.fc4inv = nn.Linear(output_dim, hidden_dim)
        self.fc3inv_ = nn.Linear(hidden_dim, hidden_dim)
        self.fc3inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc2inv_ = nn.Linear(hidden_dim, hidden_dim)
        self.fc2inv = nn.Linear(hidden_dim, hidden_dim)
        self.fc1inv = nn.Linear(hidden_dim, input_dim)

        self.relu1 = nn.ReLU()
        self.relu2 = nn.ReLU()
        self.relu2_ = nn.ReLU()
        self.relu3 = nn.ReLU()
        self.relu3_ = nn.ReLU()
        self.relu4 = nn.ReLU()

        self.PClayer1 = pc.PCLayer()
        self.PClayer2 = pc.PCLayer()
        self.PClayer2_ = pc.PCLayer()
        self.PClayer3 = pc.PCLayer()
        self.PClayer3_ = pc.PCLayer()

    def forward(self, x):
        out = self.PCimg(x)

        out = self.fc1(out)
        out = self.relu1(out)
        out = self.PClayer1(out)

        out = self.fc2(out)
        out = self.relu2(out)
        out = self.PClayer2(out)
        out = self.fc2_(out)
        out = self.relu2_(out)
        out = self.PClayer2_(out)
        
        out = self.fc3(out)
        out = self.relu3(out)
        out = self.PClayer3(out)

        out = self.fc3_(out)
        out = self.relu3_(out)
        out = self.PClayer3_(out)

        label = self.fc4(out)
        out = self.PClabel(label)

        out = self.fc4inv(out)
        out = self.PClayer3(out)
        out = self.relu3(out)
        out = self.fc3inv(out)

        out = self.PClayer3_(out)
        out = self.relu3_(out)
        out = self.fc3inv_(out)

        out = self.PClayer2(out)
        out = self.relu2(out)
        out = self.fc2inv(out)

        out = self.PClayer2_(out)
        out = self.relu2_(out)
        out = self.fc2inv_(out)

        out = self.PClayer1(out)
        out = self.relu1(out)      
        out = self.fc1inv(out)
        out = self.PCimg(out)

        return label

########################################################################
acc = [0]

total = 0.0
correct = 0.0
pred_mat_per_epoch=[]
best_accuracies = []
pred_mat=np.zeros((10,10))

T_list = [100]#10, , 1000
T_test_list = [250] #, 2500, 5000
hidden_dim_list = [6144,8000,10000]#4096, 
mode_list = [1, 3,4]#1, 
grid = [(x, y, z, q)  for x in hidden_dim_list for y in T_list for z in T_test_list for q in mode_list]
epochs = 1

for hidden_dim,num_iterations,iterations_test,mode_cnt in grid:
    train_dataset = dsets.MNIST(root='./data', 
                                        train=True, 
                                        transform=transforms.ToTensor(),
                                        download=True)

    test_dataset = dsets.MNIST(root='./data', 
                                        train=False, 
                                        transform=transforms.ToTensor(),
                                        download=True)

    train_list = []
    test_label = []
    test_list = []
    train_label = []


    if partial > 0:
        for i in range(output_dim):
            count = 0
            for j in range(len(train_dataset)):
                if  (int(train_dataset[j][1]) == i) and (count < partial):
                    image = train_dataset[j][0].view(input_dim)
                    label = torch.eye(output_dim)[train_dataset[j][1]]*3
                    train_list.append(image)
                    train_label.append(label)
                    count += 1

    if partial == 0:
        for j in range(len(train_dataset)):
            image = train_dataset[j][0].view(input_dim)
            label = torch.eye(output_dim)[train_dataset[j][1]]
            train_list.append(image)
            train_label.append(label)


    data = torch.stack(train_list)
    labels = torch.stack(train_label)


    my_dataset =  TensorDataset(data,labels)


    train_loader = DataLoader(dataset = my_dataset,
                                batch_size=batch_size,
                                shuffle = True,
                                drop_last=True)

    test_loader = DataLoader(dataset = test_dataset,
                                batch_size=test_batch_size,
                                shuffle = True,
                                drop_last=True)
    print(train_loader.dataset.tensors[0].data.shape)
    if mode_cnt == 1:
        model = Bidirectional_Model_L_1(input_dim,hidden_dim,output_dim)
        model.cuda()
    elif mode_cnt ==2:
        model = Bidirectional_Model(input_dim,hidden_dim,output_dim)
        model.cuda()
    elif mode_cnt == 3:
        model = Bidirectional_Model_L_3(input_dim,hidden_dim,output_dim)
        model.cuda()
    else:
        model = Bidirectional_Model_L_4(input_dim,hidden_dim,output_dim)
        model.cuda()

    wd =1
    pc_trainer = pc.PCTrainer(model,optimizer_x_kwargs={'lr': x_lr}, optimizer_p_kwargs={'lr': p_lr, 'weight_decay': wd}, T = num_iterations, early_stop_condition="True")# 
    prediction_trainer = pc.PCTrainer(model, optimizer_x_kwargs={'lr': xlr_test}, optimizer_p_kwargs={'lr': 0}, T = iterations_test, early_stop_condition="True")#early_stop_condition="True"
    def callback_after_t(t, pc_trainer,img,labels):
        model.PClabel.get_x().data.copy_(labels.cuda())
        model.PCimg.get_x().data.copy_(img.cuda())


    def loss_fn(output):
        return (output - labels.cuda()).pow(2).sum() * 0.5

    for epoch in range(epochs):
        print('We are currently at: Layer/epoch', mode_cnt,epoch, hidden_dim,num_iterations,iterations_test)
        model.train()
        for i, (data, labels) in enumerate(train_loader):

            callback_after_t_kwargs={'pc_trainer':pc_trainer,'img':data,'labels':labels}
            pc_trainer.train_on_batch(data.cuda(),loss_fn, is_log_progress=True,callback_after_t=callback_after_t,callback_after_t_kwargs=callback_after_t_kwargs)

        generation(prediction_trainer) 
        acc_overall,pred_matrix_per_digit = forward_acc(model.eval(),test_loader)
        acc.append(acc_overall)
        pred_mat_per_epoch.append(pred_matrix_per_digit)
        best_accuracies.append(max(acc))
        for ind in pred_mat_per_epoch:
            pred_mat+= ind.astype(int)
        acc_per_digit={}
        most_common_err={}
        pred_mat = pred_mat.astype(int)
        total_pred_mat = np.sum(pred_mat,axis=1)
        for ind in range(0,10):
            acc_per_digit[ind]=int(10000.0*pred_mat[ind][ind]/total_pred_mat[ind])/100.0
            max_error = 0
            max_index = 0
            for k in range(0,10):
                if k!=ind:
                    if pred_mat[ind][k]>max_error:
                        max_error=pred_mat[ind][k]
                        max_index = k
            most_common_err[ind]=max_index    

        print('Epoch: {}, Forward Accuracy: {}, partial: {}, batch={}, T={}, T_test = {}, hidden_dim = {}, Layer = {}'.format(epoch,acc[-1],partial,batch_size, num_iterations,iterations_test,hidden_dim,mode_cnt),file=open("bpcn_0824.txt",'a'))
    print('------ A set of parameters done! -----')    
    print("Acc and Error message of each digit:",file=open("bi_cpc_prediction_experiment.txt","a"))

    print_stuffs_per_digit=[]
    for ind in range(0,10):
        print_stuffs_per_digit.append([ind, acc_per_digit[ind],most_common_err[ind]])
    print(tabulate(print_stuffs_per_digit,headers=["Digit", "Accuracy", "Most common error"]),file=open('B_CPC_MNIST_output.txt',"a"))



